/*
 * Description: Use fadvice so a read will fill the cache
 * Author: Catalin(ux aka Dino) BOIE
 * E-mail: catab at embedromix dot ro
 */

#define __USE_GNU
#define	_GNU_SOURCE
#define __USE_XOPEN2K
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include <asm/unistd.h>
#include <fcntl.h>


void	(*old_exit)(int status) __attribute__ ((__noreturn__)) = NULL;
ssize_t	(*old_read)(int fd, void *buf, size_t count) = NULL;
ssize_t	(*old_write)(int fd, const void *buf, size_t count) = NULL;
int	(*old_close)(int fd) = NULL;
off_t	(*old_lseek)(int fd, off_t offset, int whence) = NULL;
int	(*old_closedir)(DIR *dir) = NULL;
DIR	*(*old_opendir)(const char *name) = NULL;


static short		inited = 0;
static short		debug = 0;
static off_t		noca_size = 128;	/* Default, flush at 128KiB */
static unsigned short	noca_read = 1;		/* Call fadvise for reads */
static unsigned short	noca_write = 1;		/* Call fadvise for writes */
static unsigned short	noca_ra = 1;		/* optimistic readahead */
static unsigned int	noca_sleep = 0;		/* sleep after any flush for debugging purposes, in seconds */
/*static unsigned int	dump_gen = 0;*/

struct fds_t
{
	struct fds_t	*next;
	int		fd;

	unsigned short	ra_called;
	loff_t		count;
	unsigned short	read, write;
	/* warnings */
	unsigned short	warn_dontneed:1;
	/* set */
	unsigned short	set_ra:1;
	/* File name */
	char		file[PATH_MAX];
	/* type of file */
	mode_t		type;
	unsigned short	dumped:1;
};

struct fds_t		*fds[256];


/* Prototypes */
void exit(int status) __attribute__((__noreturn__));


/* Functions */

void init(void)
{
	int	i;
	char	*x;

	if (inited == 1)
		return;

	for (i = 0; i < 256; i++)
		fds[i] = NULL;

	x = getenv("NOCA_SIZE");
	if (x != NULL)
		noca_size = strtoll(x, NULL, 10);

	x = getenv("NOCA_DEBUG");
	if (x != NULL)
		debug = strtoul(x, NULL, 10);

	x = getenv("NOCA_READ");
	if (x)
		noca_read = strtoul(x, NULL, 10);

	x = getenv("NOCA_WRITE");
	if (x)
		noca_write = strtoul(x, NULL, 10);

	x = getenv("NOCA_RA");
	if (x)
		noca_ra = strtoul(x, NULL, 10);

	x = getenv("NOCA_SLEEP");
	if (x)
		noca_sleep = strtoul(x, NULL, 10);

	if (debug >= 4)
		syslog(LOG_INFO, "noca: Flush at %zdKiB R=%d W=%d RA=%d S=%d\n",
			noca_size, noca_read, noca_write, noca_ra, noca_sleep);

	old_exit = dlsym(RTLD_NEXT, "exit");
	if (old_exit == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'exit'!\n");
		exit(1);
	}

	old_read = dlsym(RTLD_NEXT, "read");
	if (old_read == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'read'!\n");
		exit(1);
	}

	old_write = dlsym(RTLD_NEXT, "write");
	if (old_write == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'write'!\n");
		exit(1);
	}

	old_close = dlsym(RTLD_NEXT, "close");
	if (old_close == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'close'!\n");
		exit(1);
	}

	old_lseek = dlsym(RTLD_NEXT, "lseek");
	if (old_lseek == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'lseek'!\n");
		exit(1);
	}

	old_opendir = dlsym(RTLD_NEXT, "opendir");
	if (old_opendir == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'opendir'!\n");
		exit(1);
	}

	old_closedir = dlsym(RTLD_NEXT, "closedir");
	if (old_closedir == NULL) {
		syslog(LOG_ERR, "noca: Cannot resolve 'closedir'!\n");
		exit(1);
	}

	if (debug >= 5)
		syslog(LOG_INFO, "noca: Starting...\n");

	inited = 1;
}

static void fd_rm(int fd)
{
	struct fds_t	*p = NULL, *parent;

	parent = NULL;
	p = fds[fd & 0xff];
	while (p) {
		if (p->fd == fd) {
			if (parent)
				parent->next = p->next;
			else
				fds[fd & 0xff] = p->next;

				break;
			}

		parent = p;
		p = p->next;
	}

	free(p);
}

static char *fd_get_type(mode_t m)
{
	switch (m & S_IFMT) {
	case S_IFSOCK: return "sock";
	case S_IFLNK: return "link";
	case S_IFREG: return "normal";
	case S_IFBLK: return "block";
	case S_IFDIR: return "dir";
	case S_IFCHR: return "char";
	case S_IFIFO: return "fifo";
	default: return "unk";
	}
}

static struct fds_t *fd_add(int fd)
{
	struct fds_t *p, *A;
	struct stat S;
	int ret;
	mode_t m;
	char proc_path[PATH_MAX];

	/* No need to add filedescriptor if we didn't asked for */
	if ((noca_read == 0) && (noca_write == 0) && (noca_ra == 0))
		return NULL;

	ret = fstat(fd, &S);
	if (ret == -1)
		return NULL;

	m = S.st_mode & S_IFMT;
	if (!S_ISREG(m) && !S_ISDIR(m) && !S_ISBLK(m) && !S_ISLNK(m)) {
		if (debug >= 8)
			syslog(LOG_ERR, "noca: fd_add:"
				" Don't add fd %d because is"
				" not a good one (%o=%s).\n",
				fd, m, fd_get_type(m));
		return NULL;
	}

	A = (struct fds_t *) malloc(sizeof(struct fds_t));
	if (!A)
		return NULL;

	memset(A, 0, sizeof(struct fds_t));
	A->fd = fd;
	A->type = m;

	snprintf(proc_path, sizeof(proc_path), "/proc/self/fd/%d", fd);
	ret = readlink(proc_path, A->file, sizeof(A->file) - 1);
	if (ret == -1)
		strcpy(A->file, "UNKNOWN");
	else {
		A->file[ret] = '\0';
	}

	if (fds[fd & 0xff] == NULL)
		fds[fd & 0xff] = A;
	else {
		p = fds[fd & 0xff];
		while (p->next)
			p = p->next;
		p->next = A;
	}

	if (debug >= 5)
		syslog(LOG_INFO, "noca: fd_add:"
			" Added [%s] (%s) with fd %d.\n",
			A->file, fd_get_type(A->type), fd);

	return A;
}

static struct fds_t *fd_get(int fd)
{
	struct fds_t *p = NULL;

	if (fd < 0)
		return NULL;

	if (fds[fd & 0xff] != NULL) {
		p = fds[fd & 0xff];
		while (p) {
			if (p->fd == fd)
				break;
			p = p->next;
		}
	}

	if (p == NULL)
		p = fd_add(fd);

	return p;
}

static void dont_need(struct fds_t *p)
{
	int err;

	if ((noca_read == 0) && (noca_write == 0))
		return;

	if (debug >= 4)
		syslog(LOG_ERR, "noca: %s(fd=%d (%s))\n",
			__FUNCTION__, p->fd, p->file);

	/* TODO: These off/len trashes readahead?! */
	err = posix_fadvise(p->fd, 0, 0, POSIX_FADV_DONTNEED);
	if (err < 0) {
		if (p->warn_dontneed == 0)
			syslog(LOG_ERR, "noca: cannot set DONTNEED fd=%d err=[%d] [%d][%s]!\n",
				p->fd, err, errno, strerror(errno));
		p->warn_dontneed = 1;
		return;
	}

	if (noca_sleep > 0)
		sleep(noca_sleep);

	p->count = 0;
}

void exit(int status)
{
	/* Here we can dump some statistics if we want */

	old_exit(status);
}

off_t lseek(int fd, off_t offset, int whence)
{
	off_t ret;
	struct fds_t *p;
	int old_errno;

	init();

	ret = old_lseek(fd, offset, whence);
	old_errno = errno;

	/* flush current data */
	if (ret > 0) {
		p = fd_get(fd);
		if (p == NULL) {
			errno = old_errno;
			return ret;
		}

		if (debug >= 4)
			syslog(LOG_ERR, "noca: %s(fd=%d (%s), %ld, %d)\n",
				__FUNCTION__, fd, p->file, offset, whence);
	}

	errno = old_errno;
	return ret;
}

ssize_t read(int fd, void *buf, size_t count)
{
	ssize_t		ret;
	struct fds_t	*p;
	int		err;
	int old_errno;

	init();

	ret = old_read(fd, buf, count);
	old_errno = errno;

	if (noca_read == 0) {
		errno = old_errno;
		return ret;
	}

	/* is the fd in the table? */
	p = fd_get(fd);
	if (p == NULL) {
		errno = old_errno;
		return ret;
	}

	/* set readahead */
	if (p->set_ra == 0) {
		int type;

		if (debug >= 4)
			syslog(LOG_INFO, "noca: %s: setting RA on fd=%d (%s)\n",
				__FUNCTION__, fd, p->file);
		if (noca_ra > 0)
			type = POSIX_FADV_SEQUENTIAL;
		else
			type = POSIX_FADV_RANDOM;
		err = posix_fadvise(fd, 0, 0, type);
		if (err < 0) {
			syslog(LOG_ERR, "noca: cannot set SEQUENTIAL fd=%d err=[%d] [%d][%s]!\n",
				fd, err, errno, strerror(errno));
		}
		p->set_ra = 1;
	}

	if (debug >= 4)
		syslog(LOG_INFO, "noca: %s(fd=%d (%s), buf=%p, count=%zu) = %zd\n",
			__func__, fd, p->file, (void *) buf, count, ret);

	if (ret >= 0) {
		p->count += ret;

		if (p->count >= noca_size * 1024)
			dont_need(p);
	}

	errno = old_errno;
	return ret;
}

ssize_t write(int fd, const void *buf, size_t count)
{
	ssize_t		ret;
	struct fds_t	*p;
	int old_errno;

	init();

	ret = old_write(fd, buf, count);
	old_errno = errno;

	if (noca_write == 0) {
		errno = old_errno;
		return ret;
	}

	p = fd_get(fd);
	if (p == NULL)
		return ret;

	if (debug >= 4)
		syslog(LOG_INFO, "noca: %s(fd=%d (%s), buf=%p, count=%zu) = %zd\n",
			__FUNCTION__, fd, p->file, (void *) buf, count, ret);

	if (ret >= 0) {
		p->count += ret;

		if (p->count >= noca_size * 1024)
			dont_need(p);
	}

	errno = old_errno;
	return ret;
}

int close(int fd)
{
	struct fds_t *p;

	init();

	p = fd_get(fd);
	if (p == NULL)
		return old_close(fd);

	if (debug >= 4)
		syslog(LOG_INFO, "noca: %s(fd=%d (%s))\n",
			__FUNCTION__, fd, p->file);

	dont_need(p);
	fd_rm(fd);

	return old_close(fd);
}

int closedir(DIR *dir)
{
	int fd;
	struct fds_t *p;

	init();

	if (debug >= 4)
		syslog(LOG_INFO, "noca: %s(...)\n",
			__FUNCTION__);

	if (dir != NULL) {
		fd = dirfd(dir);
		p = fd_get(fd);
		if (p != NULL) {
			dont_need(p);
		}
	}

	return old_closedir(dir);
}

DIR *opendir(const char *name)
{
	init();

	if (debug >= 9)
		syslog(LOG_INFO, "noca: %s(%s)\n",
			__FUNCTION__, name);

	return old_opendir(name);
}


